/*  Project: Open Source Gmini Tools
    file: LCD.cpp, abstraction of LCD

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Emulator.h"
#include <assert.h>

#define LCD_SET_LOWER_COLUMN_ADDRESS              ((char)0x00)
#define LCD_SET_HIGHER_COLUMN_ADDRESS             ((char)0x10)
#define LCD_SET_INTERNAL_REGULATOR_RESISTOR_RATIO ((char)0x20)
#define LCD_SET_POWER_CONTROL_REGISTER            ((char)0x28)
#define LCD_SET_DISPLAY_START_LINE                ((char)0x40)
#define LCD_SET_CONTRAST_CONTROL_REGISTER         ((char)0x81)
#define LCD_SET_SEGMENT_REMAP                     ((char)0xA0)
#define LCD_SET_LCD_BIAS                          ((char)0xA2)
#define LCD_SET_ENTIRE_DISPLAY_OFF                ((char)0xA4)
#define LCD_SET_ENTIRE_DISPLAY_ON                 ((char)0xA5)
#define LCD_SET_NORMAL_DISPLAY                    ((char)0xA6)
#define LCD_SET_REVERSE_DISPLAY                   ((char)0xA7)
#define LCD_SET_MULTIPLEX_RATIO                   ((char)0xA8)
#define LCD_SET_BIAS_TC_OSC                       ((char)0xA9)
#define LCD_SET_1OVER4_BIAS_RATIO                 ((char)0xAA)
#define LCD_SET_INDICATOR_OFF                     ((char)0xAC)
#define LCD_SET_INDICATOR_ON                      ((char)0xAD)
#define LCD_SET_DISPLAY_OFF                       ((char)0xAE)
#define LCD_SET_DISPLAY_ON                        ((char)0xAF)
#define LCD_SET_PAGE_ADDRESS                      ((char)0xB0)
#define LCD_SET_COM_OUTPUT_SCAN_DIRECTION         ((char)0xC0)
#define LCD_SET_TOTAL_FRAME_PHASES                ((char)0xD2)
#define LCD_SET_DISPLAY_OFFSET                    ((char)0xD3)
#define LCD_SET_READ_MODIFY_WRITE_MODE            ((char)0xE0)
#define LCD_SOFTWARE_RESET                        ((char)0xE2)
#define LCD_NOP                                   ((char)0xE3)
#define LCD_SET_END_OF_READ_MODIFY_WRITE_MODE     ((char)0xEE)

/* LCD command codes */
#define LCD_CNTL_RESET          0xe2    /* Software reset */
#define LCD_CNTL_POWER          0x2f    /* Power control */
#define LCD_CNTL_CONTRAST       0x81    /* Contrast */
#define LCD_CNTL_OUTSCAN        0xc8    /* Output scan direction */
#define LCD_CNTL_SEGREMAP       0xa1    /* Segment remap */
#define LCD_CNTL_DISPON         0xaf    /* Display on */

#define LCD_CNTL_PAGE           0xb0    /* Page address */
#define LCD_CNTL_HIGHCOL        0x10    /* Upper column address */
#define LCD_CNTL_LOWCOL         0x00    /* Lower column address */


LCD::LCD( Emulator * gm ) {
  gmini = gm;
  for (int i = 0; i < LCDWidth; i++)
    for (int j = 0; j < LCDHeight; j++)
      mem[i][j] = 0;
  xflip = yflip = false;
};


void LCD::write( u8 reg1, u8 reg2 ) {
   u8 value = reg2;
   if( ( reg1 & 0x08 ) == 0 ) {
     if (TRACE_LCD)
       gmini->tracer.log("LCD command: 0x%02x", reg2);
      u8 arg = value & 0x0F;
      switch ( value >> 4 ) {
         case 0x0:
            curCol = ( curCol & 0xF0 ) | arg;
            break;
         case 0x1:
            curCol = ( curCol & 0x0F ) | ( arg << 4 );
            break;
         case 0xA:
	   break;
         case 0xB:
            curLine = arg;
            break;
         case 0xC:
	    assert(arg == 0 || arg == 8 || arg == 2);
	    yflip = arg != 8;
            break;
      }
   }
   else {
     if ( curCol < LCDWidth && curLine < LCDHeight ) {
       mem[curCol][curLine] = value;
       curCol++;
       updated = true;
     } else {
       console.output("BUG: lcdwrite out of area line=%x, col=%x\n", curCol, curLine);
     }
   }
}

void LCD::dump(  ) {
   for( int L = 7; L >= 0; L-- )
      for( int l = 7; l >= 0; l-- ) {
         for( int c = 0; c < 128; c++ ) {
            bool on = mem[c][L] & ( 1 << l );
            console.output( on ? "#" : " " );
         }
         console.output( "\n" );
      }
   console.output("current line=%x, col=%x\n", curCol, curLine);
}

void LCD::dumpImage(u8* buf) {
   for( int L = 7; L >= 0; L-- )
      for( int l = 7; l >= 0; l-- ) {
         for( int c = 0; c < 128; c++ ) {
            bool on = mem[c][L] & ( 1 << l );
	    if (yflip)
	      buf[c+128*(64-(L*8+l))] = on ? 1 : 0;
	    else
	      buf[c+128*(L*8+l)] = on ? 1 : 0;
         }
      }
}

