/*  Project: Open Source Gmini Tools
    file: MMU.cpp, abstraction of memory management unit

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <stdio.h>

#include <iostream>
#include <cstdlib>
#include <fstream>
#include <sstream>
using namespace std;
#include <assert.h>
#include "types.h"
#include "util.h"
#include "MMIO.h"
#include "MMU.h"
#include "CommentMap.h"
#include "Console.h"
#include "Emulator.h"

const char MMU::stateMEMfile[] = "stateMEM.bin";
const char MMU::stateRAMfile[] = "stateRAM.bin";
const char MMU::stateFIQfile[] = "stateFIQ.bin";


MMU::MMU( Emulator * gm ) {
   gmini = gm;
   flash = new u8[FLASHSIZE];
   memset( flash, 0xFF, FLASHSIZE );
   fiq = new u8[FIQSIZE];
   memset( fiq, 0, FIQSIZE );
   ram = new u8[RAMSIZE];
   // memset( ram, 0xaa, RAMSIZE );
}

MMU::~MMU(  ) {
   delete flash;
   delete fiq;
   delete ram;
}

const u16 MMU::cread16( u22 address ) {
   if( address < FLASHSIZE )
      return ( u16 ) ( ( flash[address] << 8 ) | flash[address + 1] );
   if( address >= FIQMEM && address < FIQMEM + FIQSIZE )
      return ( u16 ) ( ( fiq[address - FIQMEM] << 8 ) | fiq[address - FIQMEM + 1] );
   else
      return 0;
}

bool MMU::cwrite16( u22 address, u16 value ) {
   if( address < FLASHSIZE ) {
      flash[address + 0] = ( u8 ) ( ( value & 0xFF00 ) >> 8 );
      flash[address + 1] = ( u8 ) ( ( value & 0x00FF ) >> 0 );
      return true;
   }
   if( address >= FIQMEM && address < FIQMEM + FIQSIZE ) {
      fiq[address - FIQMEM + 0] = ( u8 ) ( ( value & 0xFF00 ) >> 8 );
      fiq[address - FIQMEM + 1] = ( u8 ) ( ( value & 0x00FF ) >> 0 );
      return true;
   }
   else
      return false;
}

const u8 MMU::dread8( u22 address ) {
   if( address < RAMSIZE )
      return ram[address];
   if( ( address > IOBASE ) && ( address < ( IOBASE + IOSIZE ) ) )
      return gmini->io.read8( address - IOBASE );
   return 0;
}

const u16 MMU::dread16( u22 address ) {
   if( address < RAMSIZE )
      return ( u16 ) ( ( ram[address] << 8 ) | ram[address + 1] );
   if( ( address > IOBASE ) && ( address < ( IOBASE + IOSIZE ) ) )
      return gmini->io.read16( address - IOBASE );
   return 0;
}


const u32 MMU::drawread32( u22 address ) {
   if( address < RAMSIZE )
      return ( u32 )( ( ram[address + 0] << 24 ) | 
                      ( ram[address + 1] << 16 ) | 
                      ( ram[address + 2] << 8  ) | 
                      ( ram[address + 3] << 0  ) );
   return 0;
}

bool MMU::dwrite8( u22 address, u8 value ) {
   if( address < RAMSIZE ) {
      ram[address] = value;
      return true;
   }
   if( ( address > IOBASE ) && ( address < ( IOBASE + IOSIZE ) ) )
      return gmini->io.write8( address - IOBASE, value );
   return false;
}

bool MMU::dwrite16( u22 address, u16 value ) {
   if( address < RAMSIZE ) {
      ram[address + 0] = ( u8 ) ( ( value & 0xFF00 ) >> 8 );
      ram[address + 1] = ( u8 ) ( ( value & 0x00FF ) >> 0 );
      return true;
   }
   if( ( address > IOBASE ) && ( address < ( IOBASE + IOSIZE ) ) )
      return gmini->io.write16( address - IOBASE, value );
   return false;
}

void MMU::drawwrite32( u22 address, u32 value ) {
   if( address < RAMSIZE ) {
      ram[address + 0] = ( u8 ) ( ( value & 0xFF000000 ) >> 24 );
      ram[address + 1] = ( u8 ) ( ( value & 0x00FF0000 ) >> 16 );
      ram[address + 2] = ( u8 ) ( ( value & 0x0000FF00 ) >> 8 );
      ram[address + 3] = ( u8 ) ( ( value & 0x000000FF ) >> 0 );
   }
}

const u16 MMU::cread( u22 address ) {
   if( address < FLASHSIZE )
      return ( u16 ) ( ( flash[address] << 8 ) | flash[address + 1] );
   if( address >= FIQMEM && address < FIQMEM + FIQSIZE )
      return ( u16 ) ( ( fiq[address - FIQMEM] << 8 ) | fiq[address - FIQMEM + 1] );
   else
      return 0;
}

bool MMU::cwrite( u22 address, u8 value ) {
   if( address < FLASHSIZE ) {
      flash[address] = value;
      return true;
   }
   if( address >= FIQMEM && address < FIQMEM + FIQSIZE ) {
      fiq[address - FIQMEM] = value;
      return true;
   }
   else
      return false;
}

void MMU::dmemcopy( u32 destination, u32 source, u32 size ) {
   for( u32 i = 0; i < size; i++ )
      ram[destination + i] = ram[source + i];
}

u8 *MMU::getdmem(  ) {
   return ram;
}

void MMU::cmemcopy( u32 destination, u32 source, u32 size ) {
   u8 *dest;
   if( destination >= FIQMEM && ( destination + size ) < ( FIQMEM + FIQSIZE ) ) {
      dest = fiq;
      destination -= FIQMEM;
   }
   else if( ( destination + size ) < FLASHSIZE )
      dest = flash;
   else
      assert( false );
   u8 *src;
   if( source >= FIQMEM && ( source + size ) < ( FIQMEM + FIQSIZE ) ) {
      src = fiq;
      source -= FIQMEM;
   }
   else if( ( source + size ) < FLASHSIZE )
      src = flash;
   else
      assert( false );          // source not in flash or fiq
   for( u32 i = 0; i < size; i++ )
      dest[destination + i] = src[source + i];
}

const u32 MMU::find( const u8 * value, u8 length ) {
   for( u32 address = 0; ( address + length ) <= FLASHSIZE; ++address )
      if( flash[address] == value[0] ) {
         u8 i = 1;
         for( ; i < length && flash[address + i] == value[i]; ++i ) ;
         if( i == length )
            return address;
      }
   return 0xFFFFFFFF;
}
const char *MMU::getcstr( char *buffer, u22 address, int maxlength ) {
   int i = 0;
   if( address < FLASHSIZE ) {
      char ch = flash[address++];
      while( i < maxlength && address < FLASHSIZE && ch != '\0' ) {
         buffer[i++] = ch;
         ch = flash[address++];
      }
   }
   buffer[i] = '\0';
   return buffer;
}
const char *MMU::getdstr( char *buffer, u22 address, int maxlength ) {
   int i = 0;
   char ch = ram[address++];
   while( i < maxlength && ch != '\0' ) {
      buffer[i++] = ch;
      ch = ram[address++];
   }
   buffer[i] = '\0';
   return buffer;
}
bool MMU::load( const char *filename ) {
   u8 *aaz;
   u32 aazlen = loadFile( filename, &aaz );
   if ( aazlen == -1U) {
     console.output("Fatal!");
     exit(-1);
   }
     
   if( bu32( aaz ) != 2 ) {
      console.output( "Unknown AAZ header version: %d\n", bu32( aaz ) );
      delete[] aaz;
      return false;
   }
   u32 flash_size = bu32( aaz + 8 );
   if( aazlen != ( flash_size + 0x100 ) ) {

     console.output( "Incorrect size. Given:%x, should be:%x\n", 
		     bu32( aaz+8 ), flash_size - 0x100 );
      delete[] aaz;
      return false;
   }
   u16 chksum = bu16( aaz + 16 );
   u16 chkchksum = chksum32( aaz + 0x100, flash_size ) & 0xffff;
   if( chksum != chkchksum ) {
      console.output( "Checksum16 error\n" );
      delete[] aaz;
      return false;
   }
   u32 address = bu32( aaz + 4 );
   bool flashHeader = bu16( aaz + 18 ) != 0;
   u22 i = address;
   if( flashHeader )
      for( int j = 0; j < 0x100; ++j )
         cwrite( i++, aaz[j] );
   for( u32 j = 0; j < flash_size; ++j )
      cwrite( i++, aaz[0x100 + j] );
   console.output( "Loaded 0x%x bytes @ 0x%x\n", flash_size, address );
   delete[] aaz;
   return true;
}

bool MMU::initialize( const char *filename ) {
   ifstream data( filename, ios::in | ios::binary );
   if( !data ) {
      console.output( "Could not open data file: %s\n", filename );
      return false;
   }
   char buf[256];
   int address = -1;
   while( !( data.eof(  ) ) ) {
      char temp[256];
      data.getline( buf, 256 );
      std::istringstream iss( buf );
      while( iss >> temp ) {
         if( ( temp[0] == '0' ) && ( temp[1] == 'x' ) ) {
            address = strtol( temp + 2, NULL, 16 );
         }
         else {
            if( address == -1 ) {
               console.output( "Address not specified\n" );
            }
            else {
               dwrite8( address++, ( char )strtol( temp, NULL, 16 ) );
            }
         }
      }
   }
   data.close(  );
   console.output( "RAM initialized with %s data.\n", filename );
   return true;
}

void MMU::findStrings(  ) {
   gmini->comments->findStrings( flash, FLASHSIZE, 0 );
}

const bool MMU::dumpmem( const char *datafile ) {
   ofstream data( datafile, ios::out | ios::binary );
   if( !data ) {
      console.output( "Could not open ROM file: %s\n", datafile );
      return false;
   }
   data.write( ( char * )flash, MEMSIZE );
   data.close(  );
   return true;
}

bool MMU::loadmem( const char *datafile ) {
   ifstream data( datafile, ios::in | ios::binary );
   if( !data ) {
      console.output( "Could not open ROM file: %s\n", datafile );
      return false;
   }
   u32 i = 0;
   u8 ch;
   while( !( data.eof(  ) ) && i < MEMSIZE ) {
      ch = data.get(  );
      flash[i++] = ch;
   }
   data.close(  );
   console.output( "Loaded %d bytes to ROM\n", i );
   return true;
}

const bool MMU::dumpfiq( const char *datafile ) {
  ofstream data;
  data.open( datafile, ios::out | ios::binary );
   if( !data ) {
      console.output( "Could not open FIQ file: %s\n", datafile );
      return false;
   }
   data.write( ( char * )fiq, FIQSIZE );
   data.close(  );
   return true;
}

bool MMU::loadfiq( const char *datafile ) {
  ifstream data;
  data.open( datafile, ios::in | ios::binary );
   if( !data ) {
      console.output( "Could not open FIQ file: %s\n", datafile );
      return false;
   }
   u32 i = 0;
   u8 ch;
   while( !( data.eof(  ) ) && i < FIQSIZE ) {
      ch = data.get(  );
      fiq[i++] = ch;
   }
   data.close(  );
   console.output( "Loaded %d bytes to FIQ\n", i );
   return true;
}

const bool MMU::dumpram( const char *datafile ) {
  ofstream data;
  data.open( datafile, ios::out | ios::binary );
   if( !data ) {
      console.output( "Could not open RAM file: %s\n", datafile );
      return false;
   }
   data.write( ( char * )ram, RAMSIZE );
   data.close(  );
   return true;
}

bool MMU::loadram( const char *datafile ) {
  ifstream data;
  data.open( datafile, ios::in | ios::binary );
   if( !data ) {
      console.output( "Could not open RAM file: %s\n", datafile );
      return false;
   }
   u32 i = 0;
   u8 ch;
   while( !( data.eof(  ) ) && i < RAMSIZE ) {
      ch = data.get(  );
      ram[i++] = ch;
   }
   data.close(  );
   console.output( "Loaded %d bytes to RAM\n", i );
   return true;
}

const void MMU::dumpScreen( u22 address, int count ) {
   int i, k;
   for( k = 0; k < count / 0x10; k++ ) {
      console.output( "%08xh  ", address + k * 16 );
      for( i = 0; i < 16; i += 2 ) {
         console.output( "%02x%02x ", ram[address + k * 16 + i], ram[address + k * 16 + i + 1] );
      }
      console.output( " " );
      for( i = 0; i < 16; i += 1 ) {
         if( isprint( ram[address + k * 16 + i] ) )
            console.output( "%c", ram[address + k * 16 + i] );
         else
            console.output( "." );
      }
      console.output( "\n" );
   }
}

#define BANK0 (0x200000)
#define BANK1 (0x220000)
#define HIGHOFS (0x10000)

const u24 MMU::xread24( u16 addr ) {
  u32 address = addr;
  u32 bank = (address&0x8000)?BANK1:BANK0;
  address &= 0x7FFF;
  u24 value =
    (((u24)ram[bank+2*address]) << 8) |
    (((u24)ram[bank+2*address+1])) |
    (((u24)ram[bank+HIGHOFS+2*address+1]) << 16);
  
//   console.output("XREAD [%x] -> %x (%x,%x,%x) -> %x,%x,%x\n",
//  		 addr, value, bank+2*address, bank+2*address+1, bank+0x10000+address,
//  		 value>>8, value, value>>16);
  return value;
}

void MMU::xwrite24( u16 addr, u24 value ) {
  u32 address = addr;
  u32 bank = (address&0x8000)?BANK1:BANK0;
  address &= 0x7FFF;
  ram[bank+2*address] = value>>8;
  ram[bank+2*address+1] = value;
  ram[bank+HIGHOFS+2*address+1] = value>>16;
//   console.output("XWRITE [%x] <- %x (%x,%x,%x) <- %x,%x,%x\n",
//  		 addr, value, bank+2*address, bank+2*address+1, bank+0x10000+address,
//  		 value>>8, value, value>>16);
}

void MMU::ydma(int dir, int blk, u16 intAddr, u22 extAddr, u16 num) {
  assert(dir == 0);
  assert(blk == 0);
  for (int i = 0; i < num; i++) {
    u24 val = 
      (((u24)ram[extAddr+3*i]) << 16) & 
      (((u24)ram[extAddr+3*i+1]) << 8) & 
      (((u24)ram[extAddr+3*i+2]) );

    xwrite24(intAddr+i, val);
  }
  
}

const void MMU::saveState(  ) {
   dumpmem( stateMEMfile );
   dumpram( stateRAMfile );
   dumpfiq( stateFIQfile );
}


void MMU::restoreState(  ) {
   loadmem( stateMEMfile );
   loadram( stateRAMfile );
   loadfiq( stateFIQfile );
}
