/*  Project: Open Source Gmini Tools
    file: util.cpp, functions to interact with the system.

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <fstream>
using namespace std;
#include <string.h>
#include "util.h"

u32 lu32( const u8 * src ) {
   return ( u32 ) ( ( *( src + 0 ) << 0 ) | ( *( src + 1 ) << 8 ) | ( *( src + 2 ) << 16 ) | ( *( src + 3 ) << 24 ) );
}

u32 bu32( const u8 * src ) {
   return ( u32 ) ( ( *( src + 3 ) << 0 ) | ( *( src + 2 ) << 8 ) | ( *( src + 1 ) << 16 ) | ( *( src + 0 ) << 24 ) );
}

u16 lu16( const u8 * src ) {
   return ( u16 ) ( ( *( src + 0 ) << 0 ) | ( *( src + 1 ) << 8 ) );
}

u16 bu16( const u8 * src ) {
   return ( u16 ) ( ( *( src + 1 ) << 0 ) | ( *( src + 0 ) << 8 ) );
}

char *lu32( char *buffer, u32 val ) {
   *( buffer + 3 ) = ( val >> 24 ) & 0xff;
   *( buffer + 2 ) = ( val >> 16 ) & 0xff;
   *( buffer + 1 ) = ( val >> 8 ) & 0xff;
   *( buffer + 0 ) = ( val >> 0 ) & 0xff;
   return buffer;
}

char *bu32( char *buffer, u32 val ) {
   *( buffer + 0 ) = ( val >> 24 ) & 0xff;
   *( buffer + 1 ) = ( val >> 16 ) & 0xff;
   *( buffer + 2 ) = ( val >> 8 ) & 0xff;
   *( buffer + 3 ) = ( val >> 0 ) & 0xff;
   return buffer;
}

char *lu16( char *buffer, u16 val ) {
   *( buffer + 1 ) = ( val >> 8 ) & 0xff;
   *( buffer + 0 ) = ( val >> 0 ) & 0xff;
   return buffer;
}

char *bu16( char *buffer, u16 val ) {
   *( buffer + 0 ) = ( val >> 8 ) & 0xff;
   *( buffer + 1 ) = ( val >> 0 ) & 0xff;
   return buffer;
}

u32 chksum32( const u8 * src, u32 len ) {
   u32 sum = 0;
   const u8 *end = src + len;
   while( src < end )
      sum = ( sum + *( src++ ) );
   return sum;
}

u8 *XOR( u8 * data, u32 length, u8 * key, u32 keylength ) {
   u32 i;
   for( i = 0; i < length; ++i )
      data[i] ^= key[i % keylength];
   return data;
}

int max( int a, int b ) {
   if( a > b )
      return a;
   return b;
}

int min( int a, int b ) {
   if( a < b )
      return a;
   return b;
}

void inflate( const u8 * src, int slen, u8 * dest, int dlen ) {
   u8 flag = 0;
   u8 used = 7;
   int si = 0;
   int di = 0;
   while( di < dlen && si < slen ) {
      flag >>= 1;
      if( ++used == 8 ) {
         flag = src[si++];
         used = 0;
      }
      if( flag & 0x01 ) {
	//printf("0 %6x %02x\n", di, src[si]);
         dest[di++] = src[si++];
      }
      else {
         int offset = ( di & 0xfffff000 ) + 18 + ( ( u32 ) src[si] | ( ( ( u32 ) src[si + 1] & 0xf0 ) << 4 ) );
         u8 count = ( src[si + 1] & 0x0f ) + 3;
         si += 2;
         if( offset > di )
            offset -= 4096;
         for( ; count > 0; --count ) {
	   //	   printf("1 %6x %02x\n", di, dest[offset]);
	   if (offset >= 0)
	     dest[di++] = dest[offset++];
	   else {
	     dest[di++] = 0x20;
	     offset++;
	     // this case is probably a workaround for Archos packer bug.
	     // -- jyp
	   }
	 }
      }
   }
}

int compress( const u8 * src, int slen, u8 * dest ) {
   int si = 0;                  // src index
   int fi = 0;                  // flag index
   int di = 0;                  // dest index
   while( si < slen ) {
      di = fi + 1;
      for( int i = 0; i < 8 && si < slen; ++i ) {
         int position = 0;
         int length = 0;
         for( int offset = max( si - 0x1000 + 0x12, 0 ); offset < si; ++offset ) {
            if( src[offset] == src[si] ) {
               int len = 1;
               while( ( si + len ) < slen && src[offset + len] == src[si + len] && len < 18 )
                  len++;
               if( len > length ) {
                  length = len;
                  position = offset - 18;
               }
            }
         }
         if( length < 3 ) {
            dest[fi] |= 1 << i;
            dest[di++] = src[si++];
         }
         else {
            dest[di++] = position & 0xff;
            dest[di++] = ( ( position >> 4 ) & 0xf0 ) | ( length - 3 );
            si += length;
         }
      }
      fi = di;
   }
   return di;
}

/* use strchr() or strrchr() instead */
int strrpos( const char *str, const char c ) {
   int i = strlen( str );
   while( i > 0 && str[i] != c )
      --i;
   return i;
}

char *cstr( char *dest, const u8 * src, u32 len ) {
   memcpy( dest, src, len );
   dest[len] = 0;
   return dest;
}

int loadFile( const char *filename, u8 ** buf ) {
   ifstream data( filename, ios::in | ios::binary | ios::ate );
   if( !data ) {
      cout << "Could not open data file: " << filename << endl;
      return -1;
   }
   u32 length = data.tellg(  );
   data.seekg( 0, ios::beg );
   u8 *buffer = new u8[length];
   data.read( ( char * )buffer, length );
   data.close(  );
   *buf = buffer;
   return length;
}

int writeFile( const char *filename, const u8 * buffer, u32 length ) {
   ofstream data( filename, ios::out | ios::binary | ios::trunc );
   if( !data ) {
      cout << "Could not open output file: " << filename << endl;
      return -1;
   }
   data.write( ( char * )buffer, length );
   data.close(  );
   return 0;
}

// allocates a char array which must be deleted.
char *basename( char *fullname ) {
   int len = strrpos( fullname, '.' );
   char *filebase = new char[len + 1];
   strncpy( filebase, fullname, len );
   filebase[len] = '\0';
   return filebase;
}

// allocates a char array which must be deleted.
char *fileextension( const char *fullname ) {
   int pos = strrpos( fullname, '.' );
   int len = strlen( fullname ) - pos;
   char *fileext = new char[len + 1];
   strncpy( fileext, fullname + pos, len );
   fileext[len] = '\0';
   return fileext;
}

char qbuffer[128];
char *quote( const char *str ) {
   int length = strlen( str );
   int i = 0;
   int j = 0;
   while( i < length ) {
      switch ( str[i] ) {
         case '>':
            qbuffer[j++] = '&';
            qbuffer[j++] = 'g';
            qbuffer[j++] = 't';
            qbuffer[j++] = ';';
            break;
         case '<':
            qbuffer[j++] = '&';
            qbuffer[j++] = 'l';
            qbuffer[j++] = 't';
            qbuffer[j++] = ';';
            break;
         default:
            qbuffer[j++] = str[i];
      }
      ++i;
   }
   qbuffer[j] = '\0';
   return qbuffer;
}
